
from __future__ import absolute_import
import zope.interface


class IKeyStoreSync(zope.interface.Interface):
    def get(key):
        """
        Fetch a key
        @return None or bytestring value.
        """

    def put(key, value):
        """
        Create or overwrite a key.
        @param key Bytestring key.
        @param value Bytestring alue.
        @return deferred producing None or error.
        """

    def delete(key):
        """
        Delete a key.
        @param key Bytestring key.
        @return deferred producing None or error.
        """

    def seek(key):
        """
        Seek to a key, or the next highest key.
        @param key Bytestring key.
        """

    def first():
        """
        Seek to the first key in the store.
        @return True if positioned on a key.
        """

    def last():
        """
        Seek to the last key in the store.
        @return True if positioned on a key.
        """

    def next():
        """
        Seek to the next key in the store.
        @return True if positioned on a key.
        """

    def prev():
        """
        Seek to the previous key in the store.
        @return True if positioned on a key.
        """


class IKeyStore(zope.interface.Interface):
    def get(key):
        """
        Fetch a key.
        @return deferred producing None or bytestring value.
        """

    def getKeys(key, count):
        """
        Fetch a list of keys.
        @param key
            Bytestring first key to return, or None for first/last key
            in space.
        @param count Number of keys including first key to return.
        """

    def getKeysReverse(key, count):
        """
        Fetch a list of keys, walking in reverse.
        @param key
            Bytestring first key to return, or None for first/last key
            in space.
        @param count Number of keys including first key to return.
        """

    def getItems(key, count):
        """
        Fetch a list of (key, value) tuples.
        @param key
            Bytestring first key to return, or None for first/last key
            in space.
        @param count Number of keys including first key to return.
        """

    def getItemsReverse(key, count):
        """
        Fetch a list of (key, value) tuples.
        @param key
            Bytestring first key to return, or None for first/last key
            in space.
        @param count Number of keys including first key to return.
        """

    def put(key):
        """
        Create or overwrite a key.

        @param key Bytestring key.
        @param value Bytestring alue.
        @return deferred producing None or error.
        """

    def delete(key):
        """
        Delete a key.
        @param key Bytestring key.
        @return deferred producing None or error.
        """

    def putGroup(func):
        """
        Execute a function in the context of synchronous (IKeyStoreSync)
        transaction, in a private thread.

        @param func Function accepting IKeyStoreSync parameter.
        @returns deferred producing None or error.
        """



#! /usr/bin/env python
#
# Copyright 2013 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.
#

# test delete(dupdata)

from __future__ import absolute_import
from __future__ import with_statement
import unittest

import testlib
from testlib import B
from testlib import BT
from testlib import KEYS, ITEMS, KEYS2, ITEMS2
from testlib import putData, putBigData


class IterationTestBase(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()  # creates 10 databases
        self.txn = self.env.begin(write=True)
        putData(self.txn)
        self.c = self.txn.cursor()
        self.empty_entry = (B(''), B(''))

    def matchList(self, ls_a, ls_b):
        return all(map(lambda x, y: x == y, ls_a, ls_b))


class IterationTestBase2(unittest.TestCase):
    """ This puts more data than its predecessor"""

    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()  # creates 10 databases
        self.txn = self.env.begin(write=True)
        putBigData(self.txn)  # HERE!
        self.c = self.txn.cursor()
        self.empty_entry = ('', '')

    def matchList(self, ls_a, ls_b):
        return all(map(lambda x, y: x == y, ls_a, ls_b))


class IterationTest(IterationTestBase):
    def testFromStart(self):
        # From start
        self.c.first()
        self.assertEqual(self.c.key(), KEYS[0])  # start of db
        test_list = [i for i in iter(self.c)]
        self.assertEqual(self.matchList(test_list, ITEMS), True)
        self.assertEqual(self.c.item(), self.empty_entry)  # end of db

    def testFromStartWithIternext(self):
        # From start with iternext
        self.c.first()
        self.assertEqual(self.c.key(), KEYS[0])  # start of db
        test_list = [i for i in self.c.iternext()]
        # remaining elements in db
        self.assertEqual(self.matchList(test_list, ITEMS), True)
        self.assertEqual(self.c.item(), self.empty_entry)  # end of db

    def testFromStartWithNext(self):
        # From start with next
        self.c.first()
        self.assertEqual(self.c.key(), KEYS[0])  # start of db
        test_list = []
        while 1:
            test_list.append(self.c.item())
            if not self.c.next():
                break
        self.assertEqual(self.matchList(test_list, ITEMS), True)

    def testFromExistentKeySetKey(self):
        self.c.first()
        self.c.set_key(KEYS[1])
        self.assertEqual(self.c.key(), KEYS[1])
        test_list = [i for i in self.c.iternext()]
        self.assertEqual(self.matchList(test_list, ITEMS[1:]), True)

    def testFromExistentKeySetRange(self):
        self.c.first()
        self.c.set_range(KEYS[1])
        self.assertEqual(self.c.key(), KEYS[1])
        test_list = [i for i in self.c.iternext()]
        self.assertEqual(self.matchList(test_list, ITEMS[1:]), True)

    def testFromNonExistentKeySetRange(self):
        self.c.first()
        self.c.set_range(B('c'))
        self.assertEqual(self.c.key(), B('d'))
        test_list = [i for i in self.c.iternext()]
        test_items = [i for i in ITEMS if i[0] > B('c')]
        self.assertEqual(self.matchList(test_list, test_items), True)

    def testFromLastKey(self):
        self.c.last()
        self.assertEqual(self.c.key(), KEYS[-1])
        test_list = [i for i in self.c.iternext()]
        self.assertEqual(self.matchList(test_list, ITEMS[-1:]), True)

    def testFromNonExistentKeyPastEnd(self):
        self.c.last()
        self.assertEqual(self.c.key(), KEYS[-1])
        # next() fails, leaving iterator in an unpositioned state.
        self.c.next()
        self.assertEqual(self.c.item(), self.empty_entry)
        # iternext() from an unpositioned state proceeds from start of DB.
        test_list = list(self.c.iternext())
        self.assertEqual(test_list, ITEMS)


class ReverseIterationTest(IterationTestBase):
    def testFromStartRev(self):
        # From start
        self.c.first()
        self.assertEqual(self.c.key(), KEYS[0])  # start of db
        test_list = [i for i in self.c.iterprev()]
        self.assertEqual(self.matchList(test_list, ITEMS[:1][::-1]), True)
        self.assertEqual(self.c.item(), self.empty_entry)  # very start of db

    def testFromExistentKeySetKeyRev(self):
        self.c.first()
        self.c.set_key(KEYS[2])
        self.assertEqual(self.c.key(), KEYS[2])
        test_list = [i for i in self.c.iterprev()]
        self.assertEqual(self.matchList(test_list, ITEMS[:3][::-1]), True)

    def testFromExistentKeySetRangeRev(self):
        self.c.first()
        self.c.set_range(KEYS[2])
        self.assertEqual(self.c.key(), KEYS[2])
        test_list = [i for i in self.c.iterprev()]
        self.assertEqual(self.matchList(test_list, ITEMS[:3][::-1]), True)

    def testFromNonExistentKeySetRangeRev(self):
        self.c.first()
        self.c.set_range(B('c'))
        self.assertEqual(self.c.key(), B('d'))
        test_list = [i for i in self.c.iterprev()]
        test_items = [i for i in ITEMS if i[0] <= B('d')]
        test_items = test_items[::-1]
        self.assertEqual(self.matchList(test_list, test_items), True)

    def testFromLastKeyRev(self):
        self.c.last()
        self.assertEqual(self.c.key(), KEYS[-1])
        test_list = [i for i in self.c.iterprev()]
        self.assertEqual(self.matchList(test_list, ITEMS[::-1]), True)

    def testFromLastKeyWithPrevRev(self):
        self.c.last()
        self.assertEqual(self.c.key(), KEYS[-1])  # end of db
        test_list = []
        while 1:
            test_list.append(self.c.item())
            if not self.c.prev():
                break
        self.assertEqual(self.matchList(test_list, ITEMS[::-1]), True)

    def testFromNonExistentKeyPastEndRev(self):
        self.c.first()
        self.assertEqual(self.c.key(), KEYS[0])
        # prev() fails, leaving iterator in an unpositioned state.
        self.c.prev()
        self.assertEqual(self.c.item(), self.empty_entry)
        # iterprev() from an unpositioned state proceeds from end of DB.
        test_list = list(self.c.iterprev())
        self.assertEqual(test_list, ITEMS[::-1])

class IterationTestWithDupsBase(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()
        db = self.env.open_db(B('db1'), dupsort=True)
        self.txn = self.env.begin(db, write=True)
        for _ in range(2):
            putData(self.txn)
        self.c = self.txn.cursor()
        self.empty_entry = ('', '')

    def matchList(self, ls_a, ls_b):
        return all(map(lambda x, y: x == y, ls_a, ls_b))


class IterationTestWithDups(IterationTestWithDupsBase):
    pass


class SeekIterationTest(IterationTestBase2):
    def testForwardIterationSeek(self):
        self.c.first()
        test_list = []
        for i in self.c.iternext():
            test_list.append(i)
            # skips d and e
            if self.c.key() == B('baa'):
                self.c.set_key(B('e'))
        test_item = [i for i in ITEMS2 if i[0] not in (B('d'), B('e'))]
        self.assertEqual(test_list, test_item)

    def testPutDuringIteration(self):
        self.c.first()
        test_list = []
        c = self.txn.cursor()
        for i in c.iternext():
            test_list.append(i)
            # adds 'i' upon seeing 'e'
            if c.key() == B('e'):
                self.c.put(B('i'), B(''))
        test_item = ITEMS2 + [(B('i'), B(''))]
        self.assertEqual(test_list, test_item)

    def testDeleteDuringIteration(self):
        self.c.first()
        test_list = []
        for i in self.c.iternext():
            # deletes 'e' upon seeing it
            if self.c.key() == B('e'):
                # Causes 'e' to be deleted, and advances cursor to next
                # element.
                self.c.delete()
                i = self.c.item()
            test_list.append(i)

        test_item = [i for i in ITEMS2 if i[0] != B('e')]
        self.assertEqual(test_list, test_item)


if __name__ == '__main__':
    unittest.main()




from pprint import pprint
import os
import shutil
import tempfile

from time import time as now
import random
import gdbm

MAP_SIZE = 1048576 * 400
DB_PATH = '/ram/testdb-gdbm'

if os.path.exists('/ram'):
    DB_PATH = '/ram/testdb-gdbm'
else:
    DB_PATH = tempfile.mktemp(prefix='dirtybench-gdbm')


def x():
    big = '' # '*' * 400

    if os.path.exists(DB_PATH):
        os.unlink(DB_PATH)

    t0 = now()
    words = set(file('/usr/share/dict/words').readlines())
    words.update([w.upper() for w in words])
    words.update([w[::-1] for w in words])
    words.update([w[::-1].upper() for w in words])
    words.update(['-'.join(w) for w in words])
    #words.update(['+'.join(w) for w in words])
    #words.update(['/'.join(w) for w in words])
    words = list(words)
    alllen = sum(len(w) for w in words)
    avglen = alllen  / len(words)
    print 'permutate %d words avglen %d took %.2fsec' % (len(words), avglen, now()-t0)

    getword = iter(words).next

    env = gdbm.open(DB_PATH, 'c')

    run = True
    t0 = now()
    last = t0
    while run:
        try:
            for _ in xrange(50000):
                word = getword()
                env[word] = big or word
        except StopIteration:
            run = False

        t1 = now()
        if (t1 - last) > 2:
            print '%.2fs (%d/sec)' % (t1-t0, len(words)/(t1-t0))
            last = t1

    t1 = now()
    print 'done all %d in %.2fs (%d/sec)' % (len(words), t1-t0, len(words)/(t1-t0))
    last = t1

    print
    print

    t0 = now()
    lst = sum(env[k] and 1 for k in env.keys())
    t1 = now()
    print 'enum %d (key, value) pairs took %.2f sec' % ((lst), t1-t0)

    t0 = now()
    lst = sum(1 or env[k] for k in reversed(env.keys()))
    t1 = now()
    print 'reverse enum %d (key, value) pairs took %.2f sec' % ((lst), t1-t0)

    t0 = now()
    for word in words:
        env[word]
    t1 = now()
    print 'rand lookup all keys %.2f sec (%d/sec)' % (t1-t0, lst/(t1-t0))

    t0 = now()
    for word in words:
        hash(env[word])
    t1 = now()
    print 'per txn rand lookup+hash all keys %.2f sec (%d/sec)' % (t1-t0, lst/(t1-t0))

    t0 = now()
    for word in words:
        hash(env[word])
    t1 = now()
    print 'rand lookup+hash all keys %.2f sec (%d/sec)' % (t1-t0, lst/(t1-t0))

    t0 = now()
    for word in words:
        env[word]
    t1 = now()
    print 'rand lookup all buffers %.2f sec (%d/sec)' % (t1-t0, lst/(t1-t0))

    t0 = now()
    for word in words:
        hash(env[word])
    t1 = now()
    print 'rand lookup+hash all buffers %.2f sec (%d/sec)' % (t1-t0, lst/(t1-t0))


    #
    # get+put
    #

    getword = iter(sorted(words)).next
    run = True
    t0 = now()
    last = t0
    while run:
        try:
            for _ in xrange(50000):
                word = getword()
                old = env[word]
                env[word] = word
        except StopIteration:
            run = False

        t1 = now()
        if (t1 - last) > 2:
            print '%.2fs (%d/sec)' % (t1-t0, len(words)/(t1-t0))
            last = t1

    t1 = now()
    print 'get+put all %d in %.2fs (%d/sec)' % (len(words), t1-t0, len(words)/(t1-t0))
    last = t1


    #
    # REPLACE
    #

    getword = iter(sorted(words)).next
    run = True
    t0 = now()
    last = t0
    while run:
        try:
            for _ in xrange(50000):
                word = getword()
                old = env[word]
        except StopIteration:
            run = False

        t1 = now()
        if (t1 - last) > 2:
            print '%.2fs (%d/sec)' % (t1-t0, len(words)/(t1-t0))
            last = t1

    t1 = now()
    print 'replace all %d in %.2fs (%d/sec)' % (len(words), t1-t0, len(words)/(t1-t0))
    last = t1




x()




import lmdb

# Open (and create if necessary) our database environment. Must specify
# max_dbs=... since we're opening subdbs.
env = lmdb.open('/tmp/address-book.lmdb', max_dbs=10)

# Now create subdbs for home and business addresses.
home_db = env.open_db(b'home')
business_db = env.open_db(b'business')


# Add some telephone numbers to each DB:
with env.begin(write=True) as txn:
    txn.put(b'mum', b'012345678', db=home_db)
    txn.put(b'dad', b'011232211', db=home_db)
    txn.put(b'dentist', b'044415121', db=home_db)
    txn.put(b'hospital', b'078126321', db=home_db)

    txn.put(b'vendor', b'0917465628', db=business_db)
    txn.put(b'customer', b'0553211232', db=business_db)
    txn.put(b'coworker', b'0147652935', db=business_db)
    txn.put(b'boss', b'0123151232', db=business_db)
    txn.put(b'manager', b'0644810485', db=business_db)


# Iterate each DB to show the keys are sorted:
with env.begin() as txn:
    for name, db in ('home', home_db), ('business', business_db):
        print('DB:', name)
        for key, value in txn.cursor(db=db):
            print('  ', key, value)
        print()


# Now let's update some phone numbers. We can specify the default subdb when
# starting the transaction, rather than pass it in every time:
with env.begin(write=True, db=home_db) as txn:
    print('Updating number for dentist')
    txn.put(b'dentist', b'099991231')

    print('Deleting number for hospital')
    txn.delete(b'hospital')
    print()

    print('Home DB is now:')
    for key, value in txn.cursor():
        print('  ', key, value)
    print()


# Now let's look up a number in the business DB
with env.begin(db=business_db) as txn:
    print('Boss telephone number:', txn.get(b'boss'))
    print()


# We got fired, time to delete all keys from the business DB.
with env.begin(write=True) as txn:
    print('Deleting all numbers from business DB:')
    txn.drop(business_db, delete=False)

    print('Adding number for recruiter to business DB')
    txn.put(b'recruiter', b'04123125324', db=business_db)

    print('Business DB is now:')
    for key, value in txn.cursor(db=business_db):
        print('  ', key, value)
    print()




from __future__ import absolute_import
import operator

import twisted.internet.defer
import twisted.internet.threads
import zope.interface

import keystore.interfaces


class LmdbKeyStoreSync(object):
    zope.interface.implements(keystore.interfaces.IKeyStoreSync)
    cursor = None

    def __init__(self, env, write):
        self.txn = env.begin(write=write)

    def __enter__(self):
        return self

    def __exit__(self, e_type, e_val, e_tb):
        if e_type:
            self.txn.abort()
        else:
            self.txn.commit()
        self.txn = None

    def get(self, key):
        return self.txn.get(key)

    def put(self, key, value):
        self.txn.put(key, value)

    def delete(self, key):
        self.txn.delete(key)

    #
    # Cursor.
    #

    @property
    def key(self):
        return self.cursor.key()

    @property
    def value(self):
        return self.cursor.value()

    def seek(self, key):
        self.cursor.set_range(key)

    def next(self):
        return self.cursor.next()

    def prev(self):
        return self.cursor.prev()


def _reader_task(env, func):
    with LmdbKeyStoreSync(env, write=False) as sync:
        return func(sync)


def _writer_task(env, func):
    with LmdbKeyStoreSync(env, write=True) as sync:
        return func(sync)


class LmdbKeyStore(object):
    zope.interface.implements(keystore.interfaces.IKeyStore)

    def __init__(self, reactor, pool, env):
        self.reactor = reactor
        self.pool = pool
        self.env = env

    def _call_in_thread(self, func):
        return twisted.internet.threads.deferToThreadPool(
            self.reactor,
            self.pool,
            func)

    def get(self, key):
        twisted.python.log.msg('get(%r, %r)', key)
        get = lambda sync: sync.get(key)
        return self._call_in_thread(lambda: _reader_task(self.env, get))

    def _get_forward(self, sync, key, count, getter):
        positioned = sync.seek(key)
        out = []
        for x in xrange(count):
            if not positioned:
                break
            out.append(getter(sync))
            positioned = sync.next()
        return out

    def _get_reverse(self, sync, key, count, getter):
        out = []
        positioned = sync.seek(key)
        if not positioned:
            positioned = sync.last()
        for x in xrange(count):
            if not positioned:
                break
            out.append(sync.key)
            positioned = sync.prev()
        return out

    _key_getter = operator.attrgetter('key')
    _item_getter = operator.attrgetter('key', 'value')

    def getKeys(self, key, count):
        get = lambda sync: self._get_forward(sync, key, count,
                                             self._key_getter)
        return self._call_in_thread(lambda: _reader_task(self.env, get))

    def getKeysReverse(self, key, count):
        get = lambda sync: self._get_reverse(sync, key, count,
                                             self._key_getter)
        return self._call_in_thread(lambda: _reader_task(self.env, get))

    def getItems(self, key, count):
        get = lambda sync: self._get_forward(sync, key, count,
                                             self._item_getter)
        return self._call_in_thread(lambda: _reader_task(self.env, get))

    def getItemsReverse(self, key, count):
        get = lambda sync: self._get_reverse(sync, key, count,
                                             self._item_getter)
        return self._call_in_thread(lambda: _reader_task(self.env, get))

    def put(self, key, value):
        twisted.python.log.msg('put(%r, %r)', key, value)
        put = lambda sync: sync.put(key, value)
        return self._call_in_thread(lambda: _writer_task(self.env, put))

    def delete(self, key):
        delete = lambda sync: sync.delete(key)
        return self._call_in_thread(lambda: _writer_task(self.env, delete))

    def putGroup(self, func):
        return self._call_in_thread(lambda: _writer_task(self.env, func))



#
# Copyright 2013 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.
#

# This is not a test suite! More like a collection of triggers for previously
# observed crashes. Want to contribute to py-lmdb? Please write a test suite!
#
# what happens when empty keys/ values passed to various funcs
# incorrect types
# try to break cpython arg parsing - too many/few/incorrect args
# Various efforts to cause Python-level leaks.
#

from __future__ import absolute_import
from __future__ import with_statement

import itertools
import random
import unittest

import lmdb
import testlib

from testlib import B
from testlib import O


class CrashTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    # Various efforts to cause segfaults.

    def setUp(self):
        self.path, self.env = testlib.temp_env()
        with self.env.begin(write=True) as txn:
            txn.put(B('dave'), B(''))
            txn.put(B('dave2'), B(''))

    def testOldCrash(self):
        txn = self.env.begin()
        dir(iter(txn.cursor()))

    def testCloseWithTxn(self):
        txn = self.env.begin(write=True)
        self.env.close()
        self.assertRaises(Exception, (lambda: list(txn.cursor())))

    def testDoubleClose(self):
        self.env.close()
        self.env.close()

    def testTxnCloseActiveIter(self):
        with self.env.begin() as txn:
            it = txn.cursor().iternext()
        self.assertRaises(Exception, (lambda: list(it)))

    def testDbCloseActiveIter(self):
        db = self.env.open_db(key=B('dave3'))
        with self.env.begin(write=True) as txn:
            txn.put(B('a'), B('b'), db=db)
            it = txn.cursor(db=db).iternext()
        self.assertRaises(Exception, (lambda: list(it)))


class IteratorTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()
        self.txn = self.env.begin(write=True)
        self.c = self.txn.cursor()

    def testEmpty(self):
        self.assertEqual([], list(self.c))
        self.assertEqual([], list(self.c.iternext()))
        self.assertEqual([], list(self.c.iterprev()))

    def testFilled(self):
        testlib.putData(self.txn)
        self.assertEqual(testlib.ITEMS, list(self.c))
        self.assertEqual(testlib.ITEMS, list(self.c))
        self.assertEqual(testlib.ITEMS, list(self.c.iternext()))
        self.assertEqual(testlib.ITEMS[::-1], list(self.txn.cursor().iterprev()))
        self.assertEqual(testlib.ITEMS[::-1], list(self.c.iterprev()))
        self.assertEqual(testlib.ITEMS, list(self.c))

    def testFilledSkipForward(self):
        testlib.putData(self.txn)
        self.c.set_range(B('b'))
        self.assertEqual(testlib.ITEMS[1:], list(self.c))

    def testFilledSkipReverse(self):
        testlib.putData(self.txn)
        self.c.set_range(B('b'))
        self.assertEqual(testlib.REV_ITEMS[-2:], list(self.c.iterprev()))

    def testFilledSkipEof(self):
        testlib.putData(self.txn)
        self.assertEqual(False, self.c.set_range(B('z')))
        self.assertEqual(testlib.REV_ITEMS, list(self.c.iterprev()))


class BigReverseTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    # Test for issue with MDB_LAST+MDB_PREV skipping chunks of database.
    def test_big_reverse(self):
        path, env = testlib.temp_env()
        txn = env.begin(write=True)
        keys = [B('%05d' % i) for i in range(0xffff)]
        for k in keys:
            txn.put(k, k, append=True)
        assert list(txn.cursor().iterprev(values=False)) == list(reversed(keys))


class MultiCursorDeleteTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def setUp(self):
        self.path, self.env = testlib.temp_env()

    def test1(self):
        """Ensure MDB_NEXT is ignored on `c1' when it was previously positioned
        on the key that `c2' just deleted."""
        txn = self.env.begin(write=True)
        cur = txn.cursor()
        while cur.first():
            cur.delete()

        for i in range(1, 10):
            cur.put(O(ord('a') + i) * i, B(''))

        c1 = txn.cursor()
        c1f = c1.iternext(values=False)
        while next(c1f) != B('ddd'):
            pass
        c2 = txn.cursor()
        assert c2.set_key(B('ddd'))
        c2.delete()
        assert next(c1f) == B('eeee')

    def test_monster(self):
        # Generate predictable sequence of sizes.
        rand = random.Random()
        rand.seed(0)

        txn = self.env.begin(write=True)
        keys = []
        for i in range(20000):
            key = B('%06x' % i)
            val = B('x' * rand.randint(76, 350))
            assert txn.put(key, val)
            keys.append(key)

        deleted = 0
        for key in txn.cursor().iternext(values=False):
            assert txn.delete(key), key
            deleted += 1

        assert deleted == len(keys), deleted


class TxnFullTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_17bf75b12eb94d9903cd62329048b146d5313bad(self):
        """
        me_txn0 previously cached MDB_TXN_ERROR permanently. Fixed by
        17bf75b12eb94d9903cd62329048b146d5313bad.
        """
        path, env = testlib.temp_env(map_size=4096 * 9, sync=False, max_spare_txns=0)
        for i in itertools.count():
            try:
                with env.begin(write=True) as txn:
                    txn.put(B(str(i)), B(str(i)))
            except lmdb.MapFullError:
                break

        # Should not crash with MDB_BAD_TXN:
        with env.begin(write=True) as txn:
            txn.delete(B('1'))


class EmptyIterTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_python3_iternext_segfault(self):
        # https://github.com/dw/py-lmdb/issues/105
        _, env = testlib.temp_env()
        txn = env.begin()
        cur = txn.cursor()
        ite = cur.iternext()
        nex = getattr(ite, 'next', getattr(ite, '__next__', None))
        assert nex is not None
        self.assertRaises(StopIteration, nex)


class MultiputTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_multiput_segfault(self):
        # http://github.com/jnwatson/py-lmdb/issues/173
        _, env = testlib.temp_env()
        db = env.open_db(b'foo', dupsort=True)
        txn = env.begin(db=db, write=True)
        txn.put(b'a', b'\x00\x00\x00\x00\x00\x00\x00\x00')
        txn.put(b'a', b'\x05')
        txn.put(b'a', b'\t')
        txn.put(b'a', b'\r')
        txn.put(b'a', b'\x11')
        txn.put(b'a', b'\x15')
        txn.put(b'a', b'\x19')
        txn.put(b'a', b'\x1d')
        txn.put(b'a', b'!')
        txn.put(b'a', b'%')
        txn.put(b'a', b')')
        txn.put(b'a', b'-')
        txn.put(b'a', b'1')
        txn.put(b'a', b'5')
        txn.commit()

class InvalidArgTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_duplicate_arg(self):
        # https://github.com/jnwatson/py-lmdb/issues/203
        _, env = testlib.temp_env()
        txn = env.begin(write=True)
        c = txn.cursor()
        self.assertRaises(TypeError, c.get, b'a', key=True)

class BadCursorTest(unittest.TestCase):
    def tearDown(self):
        testlib.cleanup()

    def test_cursor_open_failure(self):
        '''
        Test the error path for when mdb_cursor_open fails

        Note:
            this only would crash if cpython is built with Py_TRACE_REFS
        '''
        # https://github.com/jnwatson/py-lmdb/issues/216
        path, env = testlib.temp_env()
        db = env.open_db(b'db', dupsort=True)
        env.close()
        del env

        env = lmdb.open(path, readonly=True, max_dbs=4)
        txn1 = env.begin(write=False)
        db = env.open_db(b'db', dupsort=True, txn=txn1)
        txn2 = env.begin(write=False)
        self.assertRaises(lmdb.InvalidParameterError, txn2.cursor, db=db)

if __name__ == '__main__':
    unittest.main()




# Roughly approximates some of Symas microbenchmark.

from time import time
import random
import shutil
import os
import tempfile

import lmdb


val = ' ' * 100
MAX_KEYS = int(1e6)

t0 = time()

urandom = file('/dev/urandom', 'rb', 1048576).read

keys = set()
while len(keys) < MAX_KEYS:
    for _ in xrange(min(1000, MAX_KEYS - len(keys))):
        keys.add(urandom(16))

print 'make %d keys in %.2fsec' % (len(keys), time() - t0)
keys = list(keys)

if os.path.exists('/ram'):
    DB_PATH = '/ram/dbtest'
else:
    DB_PATH = tempfile.mktemp(prefix='nastybench')

if os.path.exists(DB_PATH):
    shutil.rmtree(DB_PATH)

env = lmdb.open(DB_PATH, map_size=1048576 * 1024,
    metasync=False, sync=False, map_async=True)

nextkey = iter(keys).next
run = True
while run:
    with env.begin(write=True) as txn:
        try:
            for _ in xrange(10000):
                txn.put(nextkey(), val)
        except StopIteration:
            run = False

d = time() - t0
env.sync(True)
print 'insert %d keys in %.2fsec (%d/sec)' % (len(keys), d, len(keys) / d)



nextkey = iter(keys).next
t0 = time()

with env.begin() as txn:
    try:
        while 1:
            txn.get(nextkey())
    except StopIteration:
        pass

d = time() - t0
print 'random lookup %d keys in %.2fsec (%d/sec)' % (len(keys), d, len(keys)/d)


nextkey = iter(keys).next
t0 = time()

with env.begin(buffers=True) as txn:
    try:
        while 1:
            txn.get(nextkey())
    except StopIteration:
        pass

d = time() - t0
print 'random lookup %d buffers in %.2fsec (%d/sec)' % (len(keys), d, len(keys)/d)


nextkey = iter(keys).next
t0 = time()

with env.begin(buffers=True) as txn:
    try:
        while 1:
            hash(txn.get(nextkey()))
    except StopIteration:
        pass

d = time() - t0
print 'random lookup+hash %d buffers in %.2fsec (%d/sec)' % (len(keys), d, len(keys)/d)



nextkey = iter(keys).next
t0 = time()

with env.begin(buffers=True) as txn:
    nextrec = txn.cursor().iternext().next
    try:
        while 1:
            nextrec()
    except StopIteration:
        pass

d = time() - t0
print 'seq read %d buffers in %.2fsec (%d/sec)' % (len(keys), d, len(keys)/d)




from __future__ import absolute_import
import webbrowser

import twisted.internet.reactor

import lmdb
import keystore.lmdb
import keystore.webapi


def main():
    port = 9999
    interface = '127.0.0.1'
    url = 'http://%s:%d/' % (interface, port)
    env = lmdb.open('/tmp/foo')
    reactor = twisted.internet.reactor
    pool = reactor.getThreadPool()
    store = keystore.lmdb.LmdbKeyStore(reactor, pool, env)
    site = keystore.webapi.create_site(store)
    reactor.listenTCP(port, site, interface=interface)
    reactor.callLater(0, webbrowser.open, url)
    reactor.run()

if __name__ == '__main__':
    main()



#
# Copyright 2013 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.
#

from __future__ import absolute_import
import unittest

import lmdb


class PackageExportsTest(unittest.TestCase):
    """
    Ensure the list of exported names matches a predefined list. Designed to
    ensure future interface changes to cffi.py and cpython.c don't break
    consistency of "from lmdb import *".
    """
    def test_exports(self):
        assert sorted(lmdb.__all__) == [
            'BadDbiError',
            'BadRslotError',
            'BadTxnError',
            'BadValsizeError',
            'CorruptedError',
            'Cursor',
            'CursorFullError',
            'DbsFullError',
            'DiskError',
            'Environment',
            'Error',
            'IncompatibleError',
            'InvalidError',
            'InvalidParameterError',
            'KeyExistsError',
            'LockError',
            'MapFullError',
            'MapResizedError',
            'MemoryError',
            'NotFoundError',
            'PageFullError',
            'PageNotFoundError',
            'PanicError',
            'ReadersFullError',
            'ReadonlyError',
            'TlsFullError',
            'Transaction',
            'TxnFullError',
            'VersionMismatchError',
            '_Database',
            'enable_drop_gil',
            'version',
        ]



#
# Copyright 2013-2020 The py-lmdb authors, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted only as authorized by the OpenLDAP
# Public License.
#
# A copy of this license is available in the file LICENSE in the
# top-level directory of the distribution or, alternatively, at
# <http://www.OpenLDAP.org/license.html>.
#
# OpenLDAP is a registered trademark of the OpenLDAP Foundation.
#
# Individual files and/or contributed packages may be copyright by
# other parties and/or subject to additional restrictions.
#
# This work also contains materials derived from public sources.
#
# Additional information about OpenLDAP can be obtained at
# <http://www.openldap.org/>.
#

from __future__ import absolute_import
from __future__ import with_statement

import os
import sys
import shutil
import platform

from setuptools import Extension
from setuptools import setup

try:
    import memsink
except ImportError:
    memsink = None


if hasattr(platform, 'python_implementation'):
    use_cpython = platform.python_implementation() == 'CPython'
else:
    use_cpython = True

if os.getenv('LMDB_FORCE_CFFI') is not None:
    use_cpython = False

if sys.version[:3] < '2.7' or (3, 0) < sys.version_info[:2] < (3, 4):
    sys.stderr.write('Error: py-lmdb requires at least CPython 2.7 or 3.4\n')
    raise SystemExit(1)

#
# Figure out which LMDB implementation to use.
#

if os.getenv('LMDB_INCLUDEDIR'):
    extra_include_dirs = [os.getenv('LMDB_INCLUDEDIR')]
else:
    extra_include_dirs = []

if os.getenv('LMDB_LIBDIR'):
    extra_library_dirs = [os.getenv('LMDB_LIBDIR')]
else:
    extra_library_dirs = []

extra_include_dirs += ['lib/py-lmdb']
extra_compile_args = []

patch_lmdb_source = False

if os.getenv('LMDB_FORCE_SYSTEM') is not None:
    print('py-lmdb: Using system version of liblmdb.')
    extra_sources = []
    extra_include_dirs += []
    libraries = ['lmdb']
elif os.getenv('LMDB_PURE') is not None:
    print('py-lmdb: Using bundled unmodified liblmdb; override with LMDB_FORCE_SYSTEM=1.')
    extra_sources = ['lib/mdb.c', 'lib/midl.c']
    extra_include_dirs += ['lib']
    libraries = []
else:
    print('py-lmdb: Using bundled liblmdb with py-lmdb patches; override with LMDB_FORCE_SYSTEM=1 or LMDB_PURE=1.')
    extra_sources = ['build/lib/mdb.c', 'build/lib/midl.c']
    extra_include_dirs += ['build/lib']
    extra_compile_args += ['-DHAVE_PATCHED_LMDB=1']
    libraries = []
    patch_lmdb_source = True

if patch_lmdb_source:
    if sys.platform.startswith('win'):
        try:
            import patch_ng as patch
        except ImportError:
            raise Exception('Building py-lmdb from source on Windows requires the "patch-ng" python module.')

    # Clean out any previously patched files
    dest = 'build' + os.sep + 'lib'
    try:
        os.mkdir('build')
    except Exception:
        pass

    try:
        shutil.rmtree(dest)
    except Exception:
        pass
    shutil.copytree('lib', dest)

    # Copy away the lmdb source then patch it
    if sys.platform.startswith('win'):
        patchfile = 'lib' + os.sep + 'py-lmdb' + os.sep + 'env-copy-txn.patch'
        patchset = patch.fromfile(patchfile)
        rv = patchset.apply(2, root=dest)
        if not rv:
            raise Exception('Applying patch failed')
    else:
        rv = os.system('/usr/bin/patch -N -p3 -d build/lib < lib/py-lmdb/env-copy-txn.patch')
        if rv:
            raise Exception('Applying patch failed')

# distutils perplexingly forces NDEBUG for package code!
extra_compile_args += ['-UNDEBUG']

# Disable some Clang/GCC warnings.
if not os.getenv('LMDB_MAINTAINER'):
    extra_compile_args += ['-w']


# Microsoft Visual Studio 9 ships with neither inttypes.h, stdint.h, or a sane
# definition for ssize_t, so here we add lib/win32 to the search path, which
# contains emulation header files provided by a third party. We force-include
# Python.h everywhere since it has a portable definition of ssize_t, which
# inttypes.h and stdint.h lack, and to avoid having to modify the LMDB source
# code. Advapi32 is needed for LMDB's use of Windows security APIs.
p = sys.version.find('MSC v.')
msvc_ver = int(sys.version[p + 6: p + 10]) if p != -1 else None

if sys.platform.startswith('win'):
    # If running on Visual Studio<=2010 we must provide <stdint.h>. Newer
    # versions provide it out of the box.
    if msvc_ver and not msvc_ver >= 1600:
        extra_include_dirs += ['lib\\win32-stdint']
    extra_include_dirs += ['lib\\win32']
    extra_compile_args += [r'/FIPython.h']
    libraries += ['Advapi32']


# Capture setup.py configuration for later use by cffi, otherwise the
# configuration may differ, forcing a recompile (and therefore likely compile
# errors). This happens even when `use_cpython` since user might want to
# LMDB_FORCE_CFFI=1 during testing.
with open('lmdb/_config.py', 'w') as fp:
    fp.write('CONFIG = dict(%r)\n\n' % ((
        ('extra_compile_args', extra_compile_args),
        ('extra_sources', extra_sources),
        ('extra_library_dirs', extra_library_dirs),
        ('extra_include_dirs', extra_include_dirs),
        ('libraries', libraries),
    ),))


if use_cpython:
    print('py-lmdb: Using CPython extension; override with LMDB_FORCE_CFFI=1.')
    install_requires = []
    if memsink:
        extra_compile_args += ['-DHAVE_MEMSINK',
                               '-I' + os.path.dirname(memsink.__file__)]
    ext_modules = [Extension(
        name='cpython',
        sources=['lmdb/cpython.c'] + extra_sources,
        extra_compile_args=extra_compile_args,
        libraries=libraries,
        include_dirs=extra_include_dirs,
        library_dirs=extra_library_dirs
    )]
else:
    print('Using cffi extension.')
    install_requires = ['cffi>=0.8']
    try:
        import lmdb.cffi
        ext_modules = [lmdb.cffi._ffi.verifier.get_extension()]
    except ImportError:
        sys.stderr.write('Could not import lmdb; ensure cffi is installed!\n')
        ext_modules = []

def grep_version():
    path = os.path.join(os.path.dirname(__file__), 'lmdb/__init__.py')
    with open(path) as fp:
        for line in fp:
            if line.startswith('__version__'):
                return eval(line.split()[-1])

setup(
    name='lmdb',
    version=grep_version(),
    description="Universal Python binding for the LMDB 'Lightning' Database",
    long_description="Universal Python binding for the LMDB 'Lightning' Database",
    long_description_content_type="text/plain",
    author='David Wilson',
    maintainer='Nic Watson',
    license='OpenLDAP BSD',
    url='http://github.com/jnwatson/py-lmdb/',
    packages=['lmdb'],

    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
    ],
    ext_package='lmdb',
    ext_modules=ext_modules,
    install_requires=install_requires,
)




from __future__ import absolute_import
import twisted.python.failure
import twisted.python.log
import twisted.web.http
import twisted.web.resource
import twisted.web.server


class Response(object):
    def __init__(self, body=None, headers=None, status=None):
        self.body = body
        self.headers = headers or {}
        self.status = status


class DeferrableResourceRender(object):
    def __init__(self, resource_, request):
        self.resource = resource_
        self.request = request
        self.closed = False

    def _notify_finish(self, resp):
        self.closed = resp

    def _on_render_error(self, failure):
        twisted.python.log.err(failure)
        if not self.closed:
            self.request.setResponseCode(twisted.web.http.INTERNAL_SERVER_ERROR)
            self.request.finish()

    def _on_render_done(self, resp):
        if not isinstance(resp, Response):
            exc = TypeError("render() didn't return Response, got %r" % (resp,))
            self._on_render_error(twisted.python.failure.Failure(exc))
            return

        if resp.body is not None and type(resp.body) is not bytes:
            exc = TypeError("render() returned %r, not bytes" %
                            (type(resp.body),))
            self._on_render_error(twisted.python.failure.Failure(exc))
            return

        if self.closed:
            return

        if resp.status is not None:
            self.request.setResponseCode(resp.status)
        for key, value in resp.headers.iteritems():
            self.request.setHeader(key, value)
        if resp.body is not None:
            self.request.setHeader('Content-Length', len(resp.body))
            self.request.write(resp.body)
        self.request.finish()

    def start_render(self):
        method = self.request.method
        handler = getattr(self.resource, 'render_' + method, None)
        if handler is None:
            self._on_render_done(Response(status=405))
            return

        self.request.notifyFinish().addBoth(self._notify_finish)
        d = twisted.internet.defer.maybeDeferred(handler, self.request)
        d.addCallbacks(self._on_render_done, self._on_render_error)
        return twisted.web.server.NOT_DONE_YET


class DeferrableResource(twisted.web.resource.Resource):
    def render(self, request):
        render = DeferrableResourceRender(self, request)
        return render.start_render()




from pprint import pprint
import atexit
import gzip
import itertools
import os
import shutil
import sys
import tempfile

from time import time as now
import random
import lmdb

MAP_SIZE = 1048576 * 400
DB_PATH = '/ram/testdb'
USE_SPARSE_FILES = sys.platform != 'darwin'

if os.path.exists('/ram'):
    DB_PATH = '/ram/testdb'
else:
    DB_PATH = tempfile.mktemp(prefix='dirtybench')


env = None
@atexit.register
def cleanup():
    if env:
        env.close()
    if os.path.exists(DB_PATH):
        shutil.rmtree(DB_PATH)


def reopen_env(**kwargs):
    if env:
        env.close()
    if os.path.exists(DB_PATH):
        shutil.rmtree(DB_PATH)
    return lmdb.open(DB_PATH, map_size=MAP_SIZE, writemap=USE_SPARSE_FILES, **kwargs)


def case(title, **params):
    def wrapper(func):
        t0 = now()
        count = func()
        t1 = now()
        print('%40s:  %2.3fs   %8d/sec' % (title, t1-t0, count/(t1-t0)))
        return func
    return wrapper


def x():
    big = '' # '*' * 400

    t0 = now()
    words_path = os.path.join(os.path.dirname(__file__), 'words.gz')
    words = set(gzip.open(words_path).read().splitlines())
    words.update([w.upper() for w in words])
    words.update([w[::-1] for w in words])
    words.update([w[::-1].upper() for w in words])
    words.update(['-'.join(w) for w in words])
    #words.update(['+'.join(w) for w in words])
    #words.update(['/'.join(w) for w in words])
    words = list(words)
    alllen = sum(len(w) for w in words)
    avglen = alllen  / len(words)
    print 'permutate %d words avglen %d took %.2fsec' % (len(words), avglen, now()-t0)
    print 'DB_PATH:', DB_PATH

    words_sorted = sorted(words)
    items = [(w, big or w) for w in words]
    items_sorted = [(w, big or w) for w in words_sorted]

    global env
    env = reopen_env()

    @case('insert')
    def test():
        with env.begin(write=True) as txn:
            for word in words:
                txn.put(word, big or word)
            return len(words)


    st = env.stat()
    print
    print 'stat:', st
    print 'k+v size %.2fkb avg %d, on-disk size: %.2fkb avg %d' %\
        ((2*alllen) / 1024., (2*alllen)/len(words),
         (st['psize'] * st['leaf_pages']) / 1024.,
         (st['psize'] * st['leaf_pages']) / len(words))
    print


    @case('enum (key, value) pairs')
    def test():
        with env.begin() as txn:
            return sum(1 for _ in txn.cursor())


    @case('reverse enum (key, value) pairs')
    def test():
        with env.begin() as txn:
            return sum(1 for _ in txn.cursor().iterprev())


    @case('enum (key, value) buffers')
    def test():
        with env.begin(buffers=True) as txn:
            return sum(1 for _ in txn.cursor())


    print


    @case('rand lookup')
    def test():
        with env.begin() as txn:
            for word in words:
                txn.get(word)
        return len(words)


    @case('per txn rand lookup')
    def test():
        for word in words:
            with env.begin() as txn:
                txn.get(word)
        return len(words)


    @case('rand lookup+hash')
    def test():
        with env.begin() as txn:
            for word in words:
                hash(txn.get(word))
        return len(words)


    @case('rand lookup buffers')
    def test():
        with env.begin(buffers=True) as txn:
            for word in words:
                txn.get(word)
        return len(words)


    @case('rand lookup+hash buffers')
    def test():
        with env.begin(buffers=True) as txn:
            for word in words:
                hash(txn.get(word))
        return len(words)


    @case('rand lookup buffers (cursor)')
    def test():
        with env.begin(buffers=True) as txn:
            cursget = txn.cursor().get
            for word in words:
                cursget(word)
        return len(words)


    print


    @case('get+put')
    def test():
        with env.begin(write=True) as txn:
            for word in words:
                txn.get(word)
                txn.put(word, word)
            return len(words)


    @case('replace')
    def test():
        with env.begin(write=True) as txn:
            for word in words:
                txn.replace(word, word)
        return len(words)


    @case('get+put (cursor)')
    def test():
        with env.begin(write=True) as txn:
            with txn.cursor() as cursor:
                for word in words:
                    cursor.get(word)
                    cursor.put(word, word)
            return len(words)


    @case('replace (cursor)')
    def test():
        with env.begin(write=True) as txn:
            with txn.cursor() as cursor:
                for word in words:
                    cursor.replace(word, word)
        return len(words)


    print


    env = reopen_env()
    @case('insert (rand)')
    def test():
        with env.begin(write=True) as txn:
            for word in words:
                txn.put(word, big or word)
        return len(words)


    env = reopen_env()
    @case('insert (seq)')
    def test():
        with env.begin(write=True) as txn:
            for word in words_sorted:
                txn.put(word, big or word)
        return len(words)


    env = reopen_env()
    @case('insert (rand), reuse cursor')
    def test():
        with env.begin(write=True) as txn:
            curs = txn.cursor()
            for word in words:
                curs.put(word, big or word)
        return len(words)
    env = reopen_env()


    @case('insert (seq), reuse cursor')
    def test():
        with env.begin(write=True) as txn:
            curs = txn.cursor()
            for word in words_sorted:
                curs.put(word, big or word)
        return len(words)


    env = reopen_env()
    @case('insert, putmulti')
    def test():
        with env.begin(write=True) as txn:
            txn.cursor().putmulti(items)
        return len(words)


    env = reopen_env()
    @case('insert, putmulti+generator')
    def test():
        with env.begin(write=True) as txn:
            txn.cursor().putmulti((w, big or w) for w in words)
        return len(words)


    print


    env = reopen_env()
    @case('append')
    def test():
        with env.begin(write=True) as txn:
            for word in words_sorted:
                txn.put(word, big or word, append=True)
        return len(words)


    env = reopen_env()
    @case('append, reuse cursor')
    def test():
        with env.begin(write=True) as txn:
            curs = txn.cursor()
            for word in words_sorted:
                curs.put(word, big or word, append=True)
        return len(words)


    env = reopen_env()
    @case('append+putmulti')
    def test():
        with env.begin(write=True) as txn:
            txn.cursor().putmulti(items_sorted, append=True)
        return len(words)


    print
    st = env.stat()
    print 'stat:', st
    print 'k+v size %.2fkb avg %d, on-disk size: %.2fkb avg %d' %\
        ((2*alllen) / 1024., (2*alllen)/len(words),
         (st['psize'] * st['leaf_pages']) / 1024.,
         (st['psize'] * st['leaf_pages']) / len(words))

x()




from __future__ import absolute_import
import pkg_resources
import keystore.web
import twisted.web.server


def read_resource(path):
    return pkg_resources.resource_string('keystore', path)


class KeyResource(keystore.web.DeferrableResource):
    isLeaf = True

    def __init__(self, store):
        keystore.web.DeferrableResource.__init__(self)
        self.store = store

    def _get_done(self, value):
        if value is None:
            return keystore.web.Response(status=404)
        else:
            return keystore.web.Response(body=value)

    def render_GET(self, request):
        d = self.store.get(request.path)
        d.addCallback(self._get_done)
        return d

    def render_PUT(self, request):
        value = request.content.read()
        d = self.store.put(request.path, value)
        d.addCallback(lambda _: keystore.web.Response(status=202))
        return d

    def render_DELETE(self, request):
        d = self.store.delete(request.path)
        d.addCallback(lambda _: keystore.web.Response(status=204))
        return d


class NamespaceResource(keystore.web.DeferrableResource):
    isLeaf = False

    def __init__(self, store):
        keystore.web.DeferrableResource.__init__(self)
        self.store = store

    def getChild(self, path, request):
        return KeyResource(self.store)


class StaticResource(twisted.web.resource.Resource):
    def __init__(self, pkg_path):
        twisted.web.resource.Resource.__init__(self)
        self.data = read_resource(pkg_path)

    def render(self, request):
        return self.data


def create_site(store):
    root = twisted.web.resource.Resource()
    root.putChild('', StaticResource('static/index.html'))
    root.putChild('db', NamespaceResource(store))
    return twisted.web.server.Site(root)
